@@ -25,14 +25,23 @@
 #endif
 
 #include "pygobject-private.h"
+#include "pythread.h"
 
 static PyObject *gerror_exc = NULL;
 static const gchar *pyginterface_type_id   = "PyGInterface::type";
 GQuark pyginterface_type_key  = 0;
+static const gchar *pygobject_class_init_id   = "PyGObject::class-init";
+GQuark pygobject_class_init_key  = 0;
+static const gchar *pyginterface_info_id   = "PyGInterface::info";
+GQuark pyginterface_info_key  = 0;
+
 
 static void pyg_flags_add_constants(PyObject *module, GType flags_type,
 				    const gchar *strip_prefix);
+static gboolean pyg_error_check(GError **error);
 
+static gboolean use_gil_state_api = FALSE;
+    
 /* -------------- GDK threading hooks ---------------------------- */
 
 /**
@@ -71,10 +80,11 @@ void
 pyg_destroy_notify(gpointer user_data)
 {
     PyObject *obj = (PyObject *)user_data;
+    PyGILState_STATE state;
 
-    pyg_block_threads();
+    state = pyg_gil_state_ensure();
     Py_DECREF(obj);
-    pyg_unblock_threads();
+    pyg_gil_state_release(state);
 }
 
 
@@ -95,10 +105,11 @@ static void
 pyobject_free(gpointer boxed)
 {
     PyObject *object = boxed;
+    PyGILState_STATE state;
 
-    pyg_block_threads();
+    state = pyg_gil_state_ensure();
     Py_DECREF(object);
-    pyg_unblock_threads();
+    pyg_gil_state_release(state);
 }
 
 
@@ -200,6 +211,18 @@ pyg_register_interface(PyObject *dict, const gchar *class_name,
     
 }
 
+static void
+pyg_register_interface_info(GType gtype, const GInterfaceInfo *info)
+{
+    g_type_set_qdata(gtype, pyginterface_info_key, (gpointer) info);
+}
+
+static const GInterfaceInfo *
+pyg_lookup_interface_info(GType gtype)
+{
+    return g_type_get_qdata(gtype, pyginterface_info_key);
+}
+
 /* -------------- GMainContext objects ---------------------------- */
 
 
@@ -328,14 +351,15 @@ pyg_object_set_property (GObject *object, guint property_id,
 {
     PyObject *object_wrapper, *retval;
     PyObject *py_pspec, *py_value;
+    PyGILState_STATE state;
 
-    pyg_block_threads();
+    state = pyg_gil_state_ensure();
 
     object_wrapper = pygobject_new(object);
 
     if (object_wrapper == NULL) {
-	pyg_unblock_threads();
-	g_return_if_fail(object_wrapper != NULL);
+	pyg_gil_state_release(state);
+	return;
     }
 
     py_pspec = pyg_param_spec_new(pspec);
@@ -353,7 +377,7 @@ pyg_object_set_property (GObject *object, guint property_id,
     Py_DECREF(py_pspec);
     Py_DECREF(py_value);
 
-    pyg_unblock_threads();
+    pyg_gil_state_release(state);
 }
 
 static void
@@ -362,14 +386,15 @@ pyg_object_get_property (GObject *object, guint property_id,
 {
     PyObject *object_wrapper, *retval;
     PyObject *py_pspec;
+    PyGILState_STATE state;
 
-    pyg_block_threads();
+    state = pyg_gil_state_ensure();
 
     object_wrapper = pygobject_new(object);
 
     if (object_wrapper == NULL) {
-	pyg_unblock_threads();
-	g_return_if_fail(object_wrapper != NULL);
+	pyg_gil_state_release(state);
+	return;
     }
 
     py_pspec = pyg_param_spec_new(pspec);
@@ -382,7 +407,7 @@ pyg_object_get_property (GObject *object, guint property_id,
     Py_DECREF(py_pspec);
     Py_XDECREF(retval);
     
-    pyg_unblock_threads();        
+    pyg_gil_state_release(state);
 }
 
 static void
@@ -840,6 +865,31 @@ add_properties (GType instance_type, PyObject *properties)
     return ret;
 }
 
+static void
+pyg_register_class_init(GType gtype, PyGClassInitFunc class_init)
+{
+    g_type_set_qdata(gtype, pygobject_class_init_key, class_init);
+}
+
+static int
+pyg_run_class_init(GType gtype, gpointer gclass, PyTypeObject *pyclass)
+{
+    PyGClassInitFunc class_init;
+    GType parent_type;
+    int rv;
+
+    parent_type = g_type_parent(gtype);
+    if (parent_type) {
+        rv = pyg_run_class_init(parent_type, gclass, pyclass);
+        if (rv) return rv;
+    }
+    class_init = g_type_get_qdata(gtype, pygobject_class_init_key);
+    if (class_init)
+        return class_init(gclass, pyclass);
+    return 0;
+}
+
+
 static PyObject *
 pyg_type_register(PyObject *self, PyObject *args)
 {
@@ -849,6 +899,7 @@ pyg_type_register(PyObject *self, PyObject *args)
     gchar *type_name = NULL;
     gint i, name_serial;
     GTypeQuery query;
+    gpointer gclass;
     GTypeInfo type_info = {
 	0,    /* class_size */
 
@@ -946,11 +997,9 @@ pyg_type_register(PyObject *self, PyObject *args)
 	if (!PyDict_Check(gsignals)) {
 	    PyErr_SetString(PyExc_TypeError,
 			    "__gsignals__ attribute not a dict!");
-	    Py_DECREF(gsignals);
 	    return NULL;
 	}
 	if (!add_signals(instance_type, gsignals)) {
-	    Py_DECREF(gsignals);
 	    return NULL;
 	}
 	PyDict_DelItemString(class->tp_dict, "__gsignals__");
@@ -966,11 +1015,9 @@ pyg_type_register(PyObject *self, PyObject *args)
 	if (!PyDict_Check(gproperties)) {
 	    PyErr_SetString(PyExc_TypeError,
 			    "__gproperties__ attribute not a dict!");
-	    Py_DECREF(gproperties);
 	    return NULL;
 	}
 	if (!add_properties(instance_type, gproperties)) {
-	    Py_DECREF(gproperties);
 	    return NULL;
 	}
 	PyDict_DelItemString(class->tp_dict, "__gproperties__");
@@ -979,8 +1026,38 @@ pyg_type_register(PyObject *self, PyObject *args)
 	PyErr_Clear();
     }
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    gclass = g_type_class_ref(instance_type);
+    if (pyg_run_class_init(instance_type, gclass, class)) {
+        g_type_class_unref(gclass);
+        return NULL;
+    }
+    g_type_class_unref(gclass);
+
+      /* Register interface implementations  */
+    if (class->tp_bases) {
+        for (i = 0; i < PyTuple_GET_SIZE(class->tp_bases); ++i)
+        {
+            PyTypeObject *base = (PyTypeObject *) PyTuple_GET_ITEM(class->tp_bases, i);
+            GType itype;
+            const GInterfaceInfo *iinfo;
+            
+            if (PyObject_IsSubclass((PyObject *) base,
+                                    (PyObject *) &PyGInterface_Type) != 1)
+                continue;
+            itype = pyg_type_from_object((PyObject *) base);
+            iinfo = pyg_lookup_interface_info(itype);
+            if (!iinfo) {
+                PyErr_Format(PyExc_NotImplementedError, "Interface type %s "
+                             "has no python implementation support", base->tp_name);
+                return NULL;
+            }
+            g_type_add_interface_static(instance_type, itype, iinfo);
+        }
+    } else
+        g_warning("type has no tp_bases");
+
+    Py_INCREF(class);
+    return (PyObject *) class;
 }
 
 static PyObject *
@@ -1042,8 +1119,9 @@ pyg_signal_new(PyObject *self, PyObject *args)
 }
 
 static PyObject *
-pyg_signal_list_names (PyObject *self, PyObject *args)
+pyg_signal_list_names (PyObject *self, PyObject *args, PyObject *kwargs)
 {
+    static char *kwlist[] = { "type", NULL };
     PyObject *py_itype, *list;
     GObjectClass *class;
     GType itype;
@@ -1051,7 +1129,9 @@ pyg_signal_list_names (PyObject *self, PyObject *args)
     guint *ids;
     guint i;
 
-    if (!PyArg_ParseTuple(args, "O:gobject.signal_list_names", &py_itype))
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+                                     "O:gobject.signal_list_names",
+                                     kwlist, &py_itype))
 	return NULL;
     if ((itype = pyg_type_from_object(py_itype)) == 0)
 	return NULL;
@@ -1084,6 +1164,186 @@ pyg_signal_list_names (PyObject *self, PyObject *args)
     return list;
 }
 
+static PyObject *
+pyg_signal_list_ids (PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "type", NULL };
+    PyObject *py_itype, *list;
+    GObjectClass *class;
+    GType itype;
+    guint n;
+    guint *ids;
+    guint i;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+                                     "O:gobject.signal_list_ids",
+                                     kwlist, &py_itype))
+	return NULL;
+    if ((itype = pyg_type_from_object(py_itype)) == 0)
+	return NULL;
+
+    if (!G_TYPE_IS_INSTANTIATABLE(itype) && !G_TYPE_IS_INTERFACE(itype)) {
+	PyErr_SetString(PyExc_TypeError,
+			"type must be instantiable or an interface");
+	return NULL;
+    }
+
+    class = g_type_class_ref(itype);
+    if (!class) {
+	PyErr_SetString(PyExc_RuntimeError,
+			"could not get a reference to type class");
+	return NULL;
+    }
+    ids = g_signal_list_ids(itype, &n);
+
+    list = PyTuple_New((gint)n);
+    if (list == NULL) {
+	g_free(ids);
+	g_type_class_unref(class);
+	return NULL;
+    }
+
+    for (i = 0; i < n; i++)
+	PyTuple_SetItem(list, i, PyInt_FromLong(ids[i]));
+    g_free(ids);
+    g_type_class_unref(class);
+    return list;
+}
+
+static PyObject *
+pyg_signal_lookup (PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "name", "type", NULL };
+    PyObject *py_itype;
+    GObjectClass *class;
+    GType itype;
+    gchar *signal_name;
+    guint id;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "sO:gobject.signal_lookup",
+                                     kwlist, &signal_name, &py_itype))
+	return NULL;
+    if ((itype = pyg_type_from_object(py_itype)) == 0)
+	return NULL;
+
+    if (!G_TYPE_IS_INSTANTIATABLE(itype) && !G_TYPE_IS_INTERFACE(itype)) {
+	PyErr_SetString(PyExc_TypeError,
+			"type must be instantiable or an interface");
+	return NULL;
+    }
+
+    class = g_type_class_ref(itype);
+    if (!class) {
+	PyErr_SetString(PyExc_RuntimeError,
+			"could not get a reference to type class");
+	return NULL;
+    }
+    id = g_signal_lookup(signal_name, itype);
+
+    g_type_class_unref(class);
+    return PyInt_FromLong(id);
+}
+
+static PyObject *
+pyg_signal_name (PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "signal_id", NULL };
+    const gchar *signal_name;
+    guint id;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i:gobject.signal_name",
+                                     kwlist, &id))
+	return NULL;
+    signal_name = g_signal_name(id);
+    if (signal_name)
+        return PyString_FromString(signal_name);
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+static PyObject *
+pyg_signal_query (PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist1[] = { "name", "type", NULL };
+    static char *kwlist2[] = { "signal_id", NULL };
+    PyObject *py_query, *params_list, *py_itype;
+    GObjectClass *class = NULL;
+    GType itype;
+    gchar *signal_name;
+    guint i;
+    GSignalQuery query;
+    guint id;
+
+    if (PyArg_ParseTupleAndKeywords(args, kwargs, "sO:gobject.signal_query",
+                                     kwlist1, &signal_name, &py_itype)) {
+        if ((itype = pyg_type_from_object(py_itype)) == 0)
+            return NULL;
+
+        if (!G_TYPE_IS_INSTANTIATABLE(itype) && !G_TYPE_IS_INTERFACE(itype)) {
+            PyErr_SetString(PyExc_TypeError,
+                            "type must be instantiable or an interface");
+            return NULL;
+        }
+
+        class = g_type_class_ref(itype);
+        if (!class) {
+            PyErr_SetString(PyExc_RuntimeError,
+                            "could not get a reference to type class");
+            return NULL;
+        }
+        id = g_signal_lookup(signal_name, itype);
+    } else {
+	PyErr_Clear();
+        if (!PyArg_ParseTupleAndKeywords(args, kwargs,
+                                         "i:gobject.signal_query",
+                                         kwlist2, &id)) {
+            PyErr_Clear();
+            PyErr_SetString(PyExc_TypeError,
+                            "Usage: one of:\n"
+                            "  gobject.signal_query(name, type)\n"
+                            "  gobject.signal_query(signal_id)");
+             
+	return NULL;
+        }
+    }
+
+    g_signal_query(id, &query);
+
+    if (query.signal_id == 0) {
+        Py_INCREF(Py_None);
+        py_query = Py_None;
+        goto done;
+    }
+    py_query = PyTuple_New(6);
+    if (py_query == NULL) {
+        goto done;
+    }
+    params_list = PyTuple_New(query.n_params);
+    if (params_list == NULL) {
+        Py_DECREF(py_query);
+        py_query = NULL;
+        goto done;
+    }
+
+    PyTuple_SET_ITEM(py_query, 0, PyInt_FromLong(query.signal_id));
+    PyTuple_SET_ITEM(py_query, 1, PyString_FromString(query.signal_name));
+    PyTuple_SET_ITEM(py_query, 2, pyg_type_wrapper_new(query.itype));
+    PyTuple_SET_ITEM(py_query, 3, PyInt_FromLong(query.signal_flags));
+    PyTuple_SET_ITEM(py_query, 4, pyg_type_wrapper_new(query.return_type));
+    for (i = 0; i < query.n_params; i++) {
+        PyTuple_SET_ITEM(params_list, i,
+                         pyg_type_wrapper_new(query.param_types[i]));
+    }
+    PyTuple_SET_ITEM(py_query, 5, params_list);
+
+ done:
+    if (class)
+        g_type_class_unref(class);
+
+    return py_query;
+}
+
 static PyObject *
 pyg_object_class_list_properties (PyObject *self, PyObject *args)
 {
@@ -1249,10 +1509,11 @@ handler_marshal(gpointer user_data)
 {
     PyObject *tuple, *ret;
     gboolean res;
+    PyGILState_STATE state;
 
     g_return_val_if_fail(user_data != NULL, FALSE);
 
-    pyg_block_threads();
+    state = pyg_gil_state_ensure();
 
     tuple = (PyObject *)user_data;
     ret = PyObject_CallObject(PyTuple_GetItem(tuple, 0),
@@ -1264,7 +1525,8 @@ handler_marshal(gpointer user_data)
 	res = PyObject_IsTrue(ret);
 	Py_DECREF(ret);
     }
-    pyg_unblock_threads();
+    
+    pyg_gil_state_release(state);
 
     return res;
 }
@@ -1348,12 +1610,13 @@ pyg_timeout_add(PyObject *self, PyObject *args, PyObject *kwargs)
 static gboolean
 iowatch_marshal(GIOChannel *source, GIOCondition condition, gpointer user_data)
 {
+    PyGILState_STATE state;
     PyObject *tuple, *func, *firstargs, *args, *ret;
     gboolean res;
 
     g_return_val_if_fail(user_data != NULL, FALSE);
 
-    pyg_block_threads();
+    state = pyg_gil_state_ensure();
 
     tuple = (PyObject *)user_data;
     func = PyTuple_GetItem(tuple, 0);
@@ -1373,7 +1636,7 @@ iowatch_marshal(GIOChannel *source, GIOCondition condition, gpointer user_data)
 	Py_DECREF(ret);
     }
 
-    pyg_unblock_threads();
+    pyg_gil_state_release(state);
 
     return res;
 }
@@ -1429,14 +1692,11 @@ static PyObject *
 pyg_source_remove(PyObject *self, PyObject *args)
 {
     guint tag;
-    PyObject *ret;
 
     if (!PyArg_ParseTuple(args, "i:source_remove", &tag))
 	return NULL;
 
-    ret = g_source_remove(tag) ? Py_True : Py_False;
-    Py_INCREF(ret);
-    return ret;
+    return PyBool_FromLong(g_source_remove(tag));
 }
 
 static PyObject *
@@ -1454,6 +1714,325 @@ pyg_main_context_default (PyObject *unused)
 
 }
 
+static int pyg_thread_state_tls_key = -1;
+
+/* Enable threading; note that the GIL must be held by the current
+   thread when this function is called */
+static int
+pyg_enable_threads ()
+{
+    if (getenv ("PYGTK_USE_GIL_STATE_API"))
+	use_gil_state_api = TRUE;
+
+#ifndef DISABLE_THREADING
+    if (!pygobject_api_functions.threads_enabled) {
+	PyEval_InitThreads();
+	if (!g_threads_got_initialized)
+	    g_thread_init(NULL);
+	pygobject_api_functions.threads_enabled = TRUE;
+	pyg_thread_state_tls_key = PyThread_create_key();
+    }
+    if (PYGIL_API_IS_BUGGY && !use_gil_state_api) {
+	PyThreadState* state;
+	state = PyThreadState_Get();
+	if ( state != NULL )
+	    PyThread_set_key_value(pyg_thread_state_tls_key, state);
+    }
+
+    return 0;
+#else
+    PyErr_SetString(PyExc_RuntimeError,
+                    "pygtk threading disabled at compile time");
+    return -1;
+#endif
+}
+
+static PyThreadState *
+pyg_find_thread_state (void)
+{
+    PyThreadState* state;
+    
+    if (pyg_thread_state_tls_key == -1)
+	return NULL;
+    state = PyThread_get_key_value(pyg_thread_state_tls_key);
+    if (state == NULL) {
+	state = PyGILState_GetThisThreadState();
+	if (state != NULL)
+	    PyThread_set_key_value(pyg_thread_state_tls_key, state);
+    }
+    return state;
+}
+
+static int
+pyg_gil_state_ensure_py23 (void)
+{
+    if (PYGIL_API_IS_BUGGY && !use_gil_state_api) {
+	PyThreadState* state = pyg_find_thread_state();
+
+	if (state == NULL)
+	    return PyGILState_LOCKED;
+    
+	if (state == _PyThreadState_Current)
+	    return PyGILState_LOCKED;
+	else {
+	    PyEval_RestoreThread(state);
+	    return PyGILState_UNLOCKED;
+	}
+    } else {
+	return PyGILState_Ensure();
+    }
+}
+
+static void
+pyg_gil_state_release_py23 (int flag)
+{
+    if (PYGIL_API_IS_BUGGY && !use_gil_state_api) {
+	if (flag == PyGILState_UNLOCKED) {
+	    PyThreadState* state = pyg_find_thread_state();
+	    if (state != NULL)
+		PyEval_ReleaseThread(state);
+	}
+    } else {
+	PyGILState_Release(flag);
+    }
+}
+
+static PyObject *
+pyg_threads_init (PyObject *unused, PyObject *args, PyObject *kwargs)
+{
+    if (pyg_enable_threads())
+        return NULL;
+    
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+struct _PyGChildData {
+    PyObject *func;
+    PyObject *data;
+};
+
+static void
+child_watch_func(GPid pid, gint status, gpointer data)
+{
+    struct _PyGChildData *child_data = (struct _PyGChildData *) data;
+    PyObject *retval;
+    PyGILState_STATE gil;
+
+    gil = pyg_gil_state_ensure();
+    if (child_data->data)
+        retval = PyObject_CallFunction(child_data->func, "iiO", pid, status,
+                                       child_data->data);
+    else
+        retval = PyObject_CallFunction(child_data->func, "ii", pid, status);
+    Py_XDECREF(retval);
+    pyg_gil_state_release(gil);
+}
+
+static void
+child_watch_dnotify(gpointer data)
+{
+    struct _PyGChildData *child_data = (struct _PyGChildData *) data;
+    Py_DECREF(child_data->func);
+    Py_XDECREF(child_data->data);
+    g_free(child_data);
+}
+
+
+static PyObject *
+pyg_child_watch_add(PyObject *unused, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "pid", "function", "data", "priority", NULL };
+    guint id;
+    gint priority = G_PRIORITY_DEFAULT;
+    int pid;
+    PyObject *func, *user_data = NULL;
+    struct _PyGChildData *child_data;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "iO|Oi:gobject.child_watch_add", kwlist,
+                                     &pid, &func, &user_data, &priority))
+        return NULL;
+    if (!PyCallable_Check(func)) {
+        PyErr_SetString(PyExc_TypeError,
+                        "gobject.child_watch_add: second argument must be callable");
+        return NULL;
+    }
+
+    child_data = g_new(struct _PyGChildData, 1);
+    child_data->func = func;
+    child_data->data = user_data;
+    Py_INCREF(child_data->func);
+    if (child_data->data)
+        Py_INCREF(child_data->data);
+    id = g_child_watch_add_full(priority, pid, child_watch_func,
+                                child_data, child_watch_dnotify);
+    return PyInt_FromLong(id);
+}
+
+struct _PyGChildSetupData {
+    PyObject *func;
+    PyObject *data;
+};
+
+static void
+_pyg_spawn_async_callback(gpointer user_data)
+{
+    struct _PyGChildSetupData *data;
+    PyObject *retval;
+    PyGILState_STATE gil;
+
+    data = (struct _PyGChildSetupData *) user_data;
+    gil = pyg_gil_state_ensure();
+    if (data->data)
+        retval = PyObject_CallFunction(data->func, "O", data->data);
+    else
+        retval = PyObject_CallFunction(data->func, NULL);
+    Py_XDECREF(retval);
+    Py_DECREF(data->func);
+    Py_XDECREF(data->data);
+    g_free(data);
+    pyg_gil_state_release(gil);
+}
+
+static PyObject *
+pyg_spawn_async(PyObject *unused, PyObject *args, PyObject *kwargs)
+{
+    static char *kwlist[] = { "argv", "envp", "working_directory", "flags",
+                              "child_setup", "user_data", "standard_input",
+                              "standard_output", "standard_error", NULL };
+    PyObject *pyargv, *pyenvp = NULL;
+    char **argv, **envp = NULL;
+    PyObject *func = NULL, *user_data = NULL;
+    char *working_directory = NULL;
+    int flags = 0, _stdin = -1, _stdout = -1, _stderr = -1;
+    PyObject *pystdin = NULL, *pystdout = NULL, *pystderr = NULL;
+    gint *standard_input, *standard_output, *standard_error;
+    struct _PyGChildSetupData *callback_data = NULL;
+    GError *error = NULL;
+    GPid child_pid = -1;
+    int len, i;
+
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|OsiOOOOO:gobject.spawn_async",
+                                     kwlist,
+                                     &pyargv, &pyenvp, &working_directory, &flags,
+                                     &func, &user_data,
+                                     &pystdin, &pystdout, &pystderr))
+        return NULL;
+
+    if (pystdin && PyObject_IsTrue(pystdin))
+        standard_input = &_stdin;
+    else
+        standard_input = NULL;
+
+    if (pystdout && PyObject_IsTrue(pystdout))
+        standard_output = &_stdout;
+    else
+        standard_output = NULL;
+
+    if (pystderr && PyObject_IsTrue(pystderr))
+        standard_error = &_stderr;
+    else
+        standard_error = NULL;
+
+      /* parse argv */
+    if (!PySequence_Check(pyargv)) {
+        PyErr_SetString(PyExc_TypeError,
+                        "gobject.spawn_async: first argument must be a sequence of strings");
+        return NULL;
+    }
+    len = PySequence_Length(pyargv);
+    argv = g_new0(char *, len + 1);
+    for (i = 0; i < len; ++i) {
+        PyObject *tmp = PySequence_ITEM(pyargv, i);
+        if (!PyString_Check(tmp)) {
+            PyErr_SetString(PyExc_TypeError,
+                            "gobject.spawn_async: first argument must be a sequence of strings");
+            g_free(argv);
+            Py_XDECREF(tmp);
+            return NULL;
+        }
+        argv[i] = PyString_AsString(tmp);
+        Py_DECREF(tmp);
+    }
+
+      /* parse envp */
+    if (pyenvp) {
+        if (!PySequence_Check(pyenvp)) {
+            PyErr_SetString(PyExc_TypeError,
+                            "gobject.spawn_async: second argument must be a sequence of strings");
+            g_free(argv);
+            return NULL;
+        }
+        len = PySequence_Length(pyenvp);
+        envp = g_new0(char *, len + 1);
+        for (i = 0; i < len; ++i) {
+            PyObject *tmp = PySequence_ITEM(pyenvp, i);
+            if (!PyString_Check(tmp)) {
+                PyErr_SetString(PyExc_TypeError,
+                                "gobject.spawn_async: second argument must be a sequence of strings");
+                g_free(envp);
+                Py_XDECREF(tmp);
+                return NULL;
+            }
+            envp[i] = PyString_AsString(tmp);
+            Py_DECREF(tmp);
+        }
+    }
+
+    if (func) {
+        callback_data = g_new(struct _PyGChildSetupData, 1);
+        callback_data->func = func;
+        callback_data->data = user_data;
+        Py_INCREF(callback_data->func);
+        if (callback_data->data)
+            Py_INCREF(callback_data->data);
+    }
+
+    if (!g_spawn_async_with_pipes(working_directory, argv, envp, flags,
+                                  func? _pyg_spawn_async_callback : NULL,
+                                  callback_data, &child_pid,
+                                  standard_input,
+                                  standard_output,
+                                  standard_error,
+                                  &error))
+    {
+        g_free(argv);
+        if (envp) g_free(envp);
+        if (callback_data) {
+            Py_DECREF(callback_data->func);
+            Py_XDECREF(callback_data->data);
+            g_free(callback_data);
+        }
+        pyg_error_check(&error);
+        return NULL;
+    }
+    g_free(argv);
+    if (envp) g_free(envp);
+
+    if (standard_input)
+        pystdin = PyInt_FromLong(*standard_input);
+    else {
+        Py_INCREF(Py_None);
+        pystdin = Py_None;
+    }
+
+    if (standard_output)
+        pystdout = PyInt_FromLong(*standard_output);
+    else {
+        Py_INCREF(Py_None);
+        pystdout = Py_None;
+    }
+
+    if (standard_error)
+        pystderr = PyInt_FromLong(*standard_error);
+    else {
+        Py_INCREF(Py_None);
+        pystderr = Py_None;
+    }
+
+    return Py_BuildValue("iNNN", child_pid, pystdin, pystdout, pystderr);
+}
+
 static PyMethodDef pygobject_functions[] = {
     { "type_name", pyg_type_name, METH_VARARGS },
     { "type_from_name", pyg_type_from_name, METH_VARARGS },
@@ -1463,7 +2042,11 @@ static PyMethodDef pygobject_functions[] = {
     { "type_interfaces", pyg_type_interfaces, METH_VARARGS },
     { "type_register", pyg_type_register, METH_VARARGS },
     { "signal_new", pyg_signal_new, METH_VARARGS },
-    { "signal_list_names", pyg_signal_list_names, METH_VARARGS },
+    { "signal_list_names", (PyCFunction)pyg_signal_list_names, METH_VARARGS|METH_KEYWORDS },
+    { "signal_list_ids", (PyCFunction)pyg_signal_list_ids, METH_VARARGS|METH_KEYWORDS },
+    { "signal_lookup", (PyCFunction)pyg_signal_lookup, METH_VARARGS|METH_KEYWORDS },
+    { "signal_name", (PyCFunction)pyg_signal_name, METH_VARARGS|METH_KEYWORDS },
+    { "signal_query", (PyCFunction)pyg_signal_query, METH_VARARGS|METH_KEYWORDS },
     { "list_properties", pyg_object_class_list_properties, METH_VARARGS },
     { "new", (PyCFunction)pyg_object_new, METH_VARARGS|METH_KEYWORDS },
     { "idle_add", (PyCFunction)pyg_idle_add, METH_VARARGS|METH_KEYWORDS },
@@ -1471,6 +2054,10 @@ static PyMethodDef pygobject_functions[] = {
     { "io_add_watch", (PyCFunction)pyg_io_add_watch, METH_VARARGS|METH_KEYWORDS },
     { "source_remove", pyg_source_remove, METH_VARARGS },
     { "main_context_default", (PyCFunction)pyg_main_context_default, METH_NOARGS },
+    { "threads_init", (PyCFunction)pyg_threads_init, METH_VARARGS|METH_KEYWORDS },
+    { "child_watch_add", (PyCFunction)pyg_child_watch_add, METH_VARARGS|METH_KEYWORDS },
+    { "spawn_async", (PyCFunction)pyg_spawn_async, METH_VARARGS|METH_KEYWORDS },
+
     { NULL, NULL, 0 }
 };
 
@@ -1599,13 +2186,15 @@ pyg_flags_add_constants(PyObject *module, GType flags_type,
 static gboolean
 pyg_error_check(GError **error)
 {
+    PyGILState_STATE state;
+
     g_return_val_if_fail(error != NULL, FALSE);
 
     if (*error != NULL) {
 	PyObject *exc_instance;
 	PyObject *d;
 	
-	pyg_block_threads();
+	state = pyg_gil_state_ensure();
 	
 	exc_instance = PyObject_CallFunction(gerror_exc, "z",
 					     (*error)->message);
@@ -1627,7 +2216,7 @@ pyg_error_check(GError **error)
 	Py_DECREF(exc_instance);
 	g_clear_error(error);
 	
-	pyg_unblock_threads();
+	pyg_gil_state_release(state);
 	
 	return TRUE;
     }
@@ -1639,10 +2228,13 @@ static PyObject *
 _pyg_strv_from_gvalue(const GValue *value)
 {
     gchar    **argv = (gchar **) g_value_get_boxed(value);
-    int        argc, i;
+    int        argc = 0, i;
     PyObject  *py_argv;
 
-    for (argc = -1; argv[++argc];);
+    if (argv) {
+        while (argv[argc])
+            argc++;
+    }
     py_argv = PyList_New(argc);
     for (i = 0; i < argc; ++i)
 	PyList_SET_ITEM(py_argv, i, PyString_FromString(argv[i]));
@@ -1780,7 +2372,14 @@ struct _PyGObject_Functions pygobject_api_functions = {
   
   &PyGFlags_Type,
   pyg_flags_add,
-  pyg_flags_from_gtype
+  pyg_flags_from_gtype,
+
+  FALSE, /* threads_enabled */
+  pyg_enable_threads,
+  pyg_gil_state_ensure_py23,
+  pyg_gil_state_release_py23,
+  pyg_register_class_init,
+  pyg_register_interface_info
 };
 
 #define REGISTER_TYPE(d, type, name) \
@@ -1804,15 +2403,10 @@ initgobject(void)
 
     PyGTypeWrapper_Type.ob_type = &PyType_Type;
     PyGParamSpec_Type.ob_type = &PyType_Type;
-
+    
     m = Py_InitModule("gobject", pygobject_functions);
     d = PyModule_GetDict(m);
 
-#ifdef ENABLE_PYGTK_THREADING
-    if (!g_threads_got_initialized)
-	g_thread_init(NULL);
-#endif
-
     g_type_init();
 
     PY_TYPE_OBJECT = g_boxed_type_register_static("PyObject",
@@ -1835,6 +2429,9 @@ initgobject(void)
     PyDict_SetItemString(PyGInterface_Type.tp_dict, "__gdoc__",
 			 pyg_object_descr_doc_get());
     pyginterface_type_key = g_quark_from_static_string(pyginterface_type_id);
+    pyginterface_info_key = g_quark_from_static_string(pyginterface_info_id);
+
+    pygobject_class_init_key = g_quark_from_static_string(pygobject_class_init_id);
 
     REGISTER_GTYPE(d, PyGBoxed_Type, "GBoxed", G_TYPE_BOXED);
     REGISTER_GTYPE(d, PyGPointer_Type, "GPointer", G_TYPE_POINTER); 
@@ -1892,6 +2489,15 @@ initgobject(void)
     PyModule_AddIntConstant(m, "IO_HUP",  G_IO_HUP);
     PyModule_AddIntConstant(m, "IO_NVAL", G_IO_NVAL);
 
+    PyModule_AddIntConstant(m, "SPAWN_LEAVE_DESCRIPTORS_OPEN", G_SPAWN_LEAVE_DESCRIPTORS_OPEN);
+    PyModule_AddIntConstant(m, "SPAWN_DO_NOT_REAP_CHILD", G_SPAWN_DO_NOT_REAP_CHILD);
+    PyModule_AddIntConstant(m, "SPAWN_SEARCH_PATH", G_SPAWN_SEARCH_PATH);
+    PyModule_AddIntConstant(m, "SPAWN_STDOUT_TO_DEV_NULL", G_SPAWN_STDOUT_TO_DEV_NULL);
+    PyModule_AddIntConstant(m, "SPAWN_STDERR_TO_DEV_NULL", G_SPAWN_STDERR_TO_DEV_NULL);
+    PyModule_AddIntConstant(m, "SPAWN_CHILD_INHERITS_STDIN", G_SPAWN_CHILD_INHERITS_STDIN);
+    PyModule_AddIntConstant(m, "SPAWN_FILE_AND_ARGV_ZERO", G_SPAWN_FILE_AND_ARGV_ZERO);
+
+
     PyModule_AddObject(m, "TYPE_INVALID", pyg_type_wrapper_new(G_TYPE_INVALID));
     PyModule_AddObject(m, "TYPE_NONE", pyg_type_wrapper_new(G_TYPE_NONE));
     PyModule_AddObject(m, "TYPE_INTERFACE", pyg_type_wrapper_new(G_TYPE_INTERFACE));
