@@ -88,8 +88,12 @@ pyg_flags_repr(PyGFlags *self)
   
     tmp = generate_repr(self->gtype, self->parent.ob_ival);
 
-    retval = g_strdup_printf("<flags %s of type %s>", tmp,
-			     g_type_name(self->gtype));
+    if (tmp)
+        retval = g_strdup_printf("<flags %s of type %s>", tmp,
+                                 g_type_name(self->gtype));
+    else
+        retval = g_strdup_printf("<flags %ld of type %s>", self->parent.ob_ival,
+                                 g_type_name(self->gtype));
     g_free(tmp);
     
     pyretval = PyString_FromString(retval);
